from abc import ABC, abstractmethod

from loguru import logger

from recurvedata.agent._internal import host, worker
from recurvedata.agent._internal.schemas import (
    CancelPayload,
    Heartbeat,
    SourceCodeExecPayload,
    TaskStatus,
    WorkerManagementPayload,
)
from recurvedata.agent._internal.task_manager import global_task_manager
from recurvedata.agent._internal.websocket.connector import WebSocketConnector
from recurvedata.agent._internal.websocket.schemas import MessageType, WebSocketMessage
from recurvedata.agent.dpserver.client import request_dpserver
from recurvedata.agent.dpserver.schema import DPServerRequestPayload

from ..worker_management import WorkerManagementTask


class BaseMessageHandler(ABC):
    _bind_message_type = None

    @abstractmethod
    async def handle(self, message: WebSocketMessage, ws_connector: WebSocketConnector):
        raise NotImplementedError


class MessageHandlerRegistry:
    def __init__(self):
        self.handlers = {}

    def register(self, message_type: MessageType, handler: BaseMessageHandler):
        self.handlers[message_type] = handler

    def register_handler(self):
        """
        Decorator: Automatically register handler to registry

        Args:
            message_type: Message type used as registry key
        """

        def decorator(handler_class):
            # Ensure handler_class is a subclass of BaseMessageHandler
            if not issubclass(handler_class, BaseMessageHandler):
                raise ValueError(f"Handler class {handler_class.__name__} must inherit from BaseMessageHandler")

            message_type = handler_class._bind_message_type
            # check in development stage
            assert message_type in MessageType, f"Register Message Handler Error, unknown message type: {message_type}"

            # Register to registry
            self.register(message_type, handler_class())

            return handler_class

        return decorator

    def get_handler(self, message_type: MessageType | None) -> BaseMessageHandler | None:
        """Get handler for specified message type"""
        return self.handlers.get(message_type)

    def get_all_handlers(self) -> dict:
        """Get all registered handlers"""
        return self.handlers.copy()

    def has_handler(self, message_type: str | None) -> bool:
        """Check if handler is registered for specified message type"""
        return message_type in self.handlers


# global registry instance
registry = MessageHandlerRegistry()


@registry.register_handler()
class HeartBeatRequestMessageHandler(BaseMessageHandler):
    """
    Handle heartbeat request message from server.
    Reply with heartbeat message to server.
    """

    _bind_message_type = MessageType.HEARTBEAT_REQUEST

    async def handle(self, message: WebSocketMessage, ws_connector: WebSocketConnector):
        logger.info(f"received heartbeat request: {message}")
        path = worker.get_docker_root_dir()
        payload = Heartbeat(
            agent_id=ws_connector.config.agent_id,
            metrics=host.get_host_metrics(path=path),
            service_status=worker.get_service_status(),
        )
        message = WebSocketMessage(
            type=MessageType.HEARTBEAT,
            payload=payload.model_dump(mode="json"),
            reply_to=message.message_id,  # reply to the same message id
        )
        await ws_connector.send_json(message.model_dump(mode="json"))


@registry.register_handler()
class WorkerManagementMessageHandler(BaseMessageHandler):
    _bind_message_type = MessageType.WORKER_MANAGEMENT

    async def handle(self, message: WebSocketMessage, ws_connector: WebSocketConnector):
        logger.info(f"received worker management message: {message.type}-{message.message_id}")
        payload = WorkerManagementPayload.model_validate(
            {
                "action": message.payload.get("action"),
                # inner payload is task payload
                "payload": message.payload,
            }
        )
        result = await WorkerManagementTask.handle(payload)
        response_message = WebSocketMessage(
            type=MessageType.RESULT,
            payload={
                "result": result.model_dump(mode="json") if result else None,
                "error": None,
                "status": TaskStatus.SUCCESS,
            },
            reply_to=message.message_id,
            message_id=message.message_id,
        )
        await ws_connector.send_json(response_message.model_dump(mode="json"))
        return result


@registry.register_handler()
class CancelMessageHandler(BaseMessageHandler):
    _bind_message_type = MessageType.CANCEL

    async def handle(self, message: WebSocketMessage, ws_connector: WebSocketConnector):
        logger.info(f"received cancel message: {message.type}-{message.message_id}")
        task_id = CancelPayload.model_validate(message.payload).task_id
        await global_task_manager.cancel_task(task_id)


@registry.register_handler()
class BusinessMessageHandler(BaseMessageHandler):
    """
    Handle business message from server.
    Forward dpserver requests to dpserver and return the response.
    """

    _bind_message_type = MessageType.BUSINESS

    async def handle(self, message: WebSocketMessage, ws_connector: WebSocketConnector):
        logger.info(f"received business message: {message.type}-{message.message_id}")

        # Parse request payload
        payload = DPServerRequestPayload.model_validate(message.payload)

        # Call dpserver to process request
        result = await request_dpserver(payload)

        # Send success response, DPServer has nothing need to be sent as log
        response_message = WebSocketMessage(
            type=MessageType.RESULT,
            payload={
                "result": result.model_dump() if result else None,
                "error": result.error if result else None,
                "status": TaskStatus.SUCCESS if result.ok else TaskStatus.FAILED,
            },
            reply_to=message.message_id,
            message_id=message.message_id,
        )
        await ws_connector.send_json(response_message.model_dump(mode="json"))
        # return for logger
        return result


@registry.register_handler()
class SourceCodeExecMessageHandler(BaseMessageHandler):
    _bind_message_type = MessageType.SOURCE_CODE_EXEC

    async def handle(self, message: WebSocketMessage, ws_connector: WebSocketConnector):
        from recurvedata.agent._internal.task_executor import TaskExecutor

        logger.info(f"received source code exec message: {message.type}-{message.message_id}")
        payload = SourceCodeExecPayload.model_validate(message.payload)
        task_executor = TaskExecutor()
        result = await task_executor.execute(payload)

        response_message = WebSocketMessage(
            type=MessageType.RESULT,
            payload=result.model_dump(),
            reply_to=message.message_id,
            message_id=message.message_id,
        )
        await ws_connector.send_json(response_message.model_dump(mode="json"))

        return result



import asyncio
import base64
import functools
import json
import logging
import os
import signal
import subprocess
import sys
import tempfile
import time
import traceback
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from loguru import logger

from ..config import RECURVE_HOME
from .schemas import SourceCodeExecPayload, TaskResultPayload, TaskStatus
from .utils import utcnow

if TYPE_CHECKING:
    from logging import Logger


@dataclass
class TaskExecutor:
    task_heartbeat_interval: float = 1
    task_retry_interval: int = 5
    subprocess_pid: int | None = None

    async def execute(self, task: "SourceCodeExecPayload") -> "TaskResultPayload":
        start_time = time.time()
        with tempfile.TemporaryDirectory() as temp_dir:
            logger.info(f"Task {task.task_instance_id} started in working directory: {temp_dir}")
            result = await self._execute_in_tempdir(task, temp_dir)

        logger.info(f"Task {task.task_instance_id} completed in {time.time() - start_time:.2f} seconds.")
        return result

    async def _execute_in_tempdir(self, task: "SourceCodeExecPayload", workdir: str) -> "TaskResultPayload":
        task_instance_id = task.task_instance_id

        cmd_args, output_file = self._prepare_execution_context(task, workdir)

        log_queue = asyncio.Queue()
        executor_logger = _ExecutorLogger(log_queue)

        try:
            await executor_logger.info(f"Executing task: {task_instance_id}")

            task_logger = _setup_task_logger(task_instance_id)
            await self.execute_in_subprocess(
                *cmd_args,
                cwd=workdir,
                log_queue=log_queue,
                timeout=task.max_duration,
                task_logger=task_logger,
            )

            # success, read result and update status
            result = None
            if output_file.exists():
                result = json.loads(output_file.read_text())

            await executor_logger.info(f"Task {task_instance_id} completed successfully.")

            return TaskResultPayload(
                status=TaskStatus.SUCCESS,
                result=result,
                logs=await self.drain_queue(log_queue),
            )
        except (subprocess.CalledProcessError, TimeoutError) as e:
            await executor_logger.error(f"Task {task_instance_id} failed: {e}")
            # raise for outside retry
            raise
        except Exception as e:
            await executor_logger.error(f"Task {task_instance_id} failed: {e}")
            return TaskResultPayload(
                status=TaskStatus.FAILED,
                error={"reason": str(e), "traceback": traceback.format_exc()},
                logs=await self.drain_queue(log_queue),
            )

    @staticmethod
    def _prepare_execution_context(task: "SourceCodeExecPayload", workdir: str) -> tuple[tuple[str, ...], Path]:
        script = Path(workdir) / f"handler.{task.handler_format}"
        script.write_bytes(base64.b64decode(task.handler_code))

        input_file = Path(workdir) / "payload.json"
        input_file.write_text(json.dumps(task.payload, indent=2))

        output_file = Path(workdir) / "result.json"

        # python /path/to/handler.py --input /path/to/payload.json --output /path/to/result.json
        cmd_args = (
            str(script),
            "--input",
            str(input_file),
            "--output",
            str(output_file),
        )
        return cmd_args, output_file

    async def execute_in_subprocess(
        self,
        *args: str,
        cwd: str,
        log_queue: asyncio.Queue[str],
        timeout: int,
        task_logger: "Logger",
    ):
        env = os.environ.copy()
        env["RECURVE_HOME"] = RECURVE_HOME

        process = await asyncio.create_subprocess_exec(
            sys.executable,
            *args,
            cwd=cwd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=env,
        )
        self.subprocess_pid = process.pid
        try:
            await asyncio.wait_for(
                asyncio.gather(
                    self.read_stream(process.stdout, log_queue, task_logger),
                    self.read_stream(process.stderr, log_queue, task_logger),
                ),
                timeout=timeout,
            )
        except asyncio.TimeoutError:
            logger.error(f"Subprocess timed out after {timeout} seconds, killing it.")
            process.kill()
            await process.wait()
            raise TimeoutError(f"Subprocess timed out after {timeout} seconds")

        rc = await process.wait()
        if rc < 0:
            signal_number = -rc
            signal_name = signal.Signals(signal_number).name
            msg = f"Subprocess was terminated by signal: {signal_name} (signal number: {signal_number})"
            logger.warning(msg)
            await log_queue.put(_format_log_message(msg))
        elif rc != 0:
            raise subprocess.CalledProcessError(rc, "python")

    @staticmethod
    async def read_stream(stream: asyncio.StreamReader, log_queue: asyncio.Queue[str], task_logger: "Logger"):
        while True:
            line = await stream.readline()
            if not line:
                break
            log_message = line.decode().strip()
            await log_queue.put(_format_log_message(log_message))
            task_logger.info(log_message)  # Print the log message to the console

    @staticmethod
    async def drain_queue(log_queue: asyncio.Queue[str]) -> list[str]:
        logs: list[str] = []
        while not log_queue.empty():
            logs.append(await log_queue.get())
        return logs


@dataclass
class _ExecutorLogger:
    log_queue: asyncio.Queue[str]

    async def info(self, msg: str):
        logger.info(msg)
        await self.log_queue.put(_format_log_message(msg))

    async def warning(self, msg: str):
        logger.warning(msg)
        await self.log_queue.put(_format_log_message(msg))

    async def error(self, msg: str):
        logger.error(msg)
        await self.log_queue.put(_format_log_message(msg))


def _format_log_message(msg: str) -> str:
    return f"{utcnow()} | {msg}"


@functools.lru_cache(maxsize=20)
def _setup_task_logger(task_id: int) -> "Logger":
    task_logger = logging.getLogger(f"task-{task_id}")
    task_logger.handlers = []
    task_logger.setLevel(logging.INFO)
    handler = logging.StreamHandler()
    formatter = logging.Formatter("[%(name)s] | %(message)s")
    handler.setFormatter(formatter)
    task_logger.addHandler(handler)
    return task_logger



import json
import time
from datetime import UTC, datetime
from typing import Any, Callable, Coroutine

import tenacity
import websockets
from loguru import logger
from websockets.exceptions import ConnectionClosed, WebSocketException

from recurvedata.agent._version import VERSION
from recurvedata.agent.config import AgentConfig
from recurvedata.agent.exceptions import UnauthorizedError

KiB = 1024
MiB = 1024 * KiB
GiB = 1024 * MiB


class RetryableWebSocketException(Exception):
    pass


class WebSocketConnector:
    """Simple async WebSocket connector for agent communication"""

    def __init__(self, config: AgentConfig):
        self.config = config
        self.ws: websockets.ClientConnection | None = None

        # Callback functions
        self.on_receive_message: Callable[[str, "WebSocketConnector"], Coroutine[Any, Any, None]] | None = None
        self.on_connect: Callable[[], Coroutine[Any, Any, None]] | None = None
        self.on_disconnect: Callable[[], Coroutine[Any, Any, None]] | None = None
        self.on_error: Callable[[Exception], Coroutine[Any, Any, None]] | None = None

        # Logger
        self.logger = logger

    @property
    def is_connected(self) -> bool:
        """Check if WebSocket is connected"""
        return self.ws is not None

    def _get_websocket_url(self) -> str:
        """Build WebSocket URL"""
        base_url = self.config.websocket_url
        return f"{base_url}/ws/"

    def _get_headers(self) -> dict:
        """Get authentication headers"""
        return {
            "Authorization": f"Bearer {self.config.agent_id}:{self.config.token.get_secret_value()}",
            "X-Tenant-Domain": self.config.tenant_domain,
            "User-Agent": f"RecurveAgent/{VERSION}",
        }

    async def connect(self) -> bool:
        """Connect to WebSocket server"""
        if self.is_connected:
            return True

        self.logger.info("Connecting to WebSocket server...")

        try:
            url = self._get_websocket_url()
            headers = self._get_headers()

            self.ws = await websockets.connect(
                url,
                additional_headers=headers,
                # Auto ping for keepalive
                ping_interval=10,
                ping_timeout=8,
                max_size=10 * MiB,
            )

            self._last_activity = time.time()
            self.logger.info("WebSocket connected successfully")

            if self.on_connect:
                try:
                    await self.on_connect()
                except Exception as e:
                    self.logger.error(f"Connect callback failed: {e}")

            return True

        except Exception as e:
            self.logger.error(f"Failed to connect: {e}")
            self.ws = None
            if self.on_error:
                try:
                    await self.on_error(e)
                except Exception as callback_error:
                    self.logger.error(f"Error callback failed: {callback_error}")
                    raise callback_error from e

            raise

    async def disconnect(self) -> None:
        """Disconnect from WebSocket server"""
        if self.is_connected:
            self.logger.info("Disconnecting from WebSocket server...")
            try:
                await self.ws.close()
            except Exception as e:
                self.logger.debug(f"Error closing WebSocket (might be already closed): {e}")

        # Always release the instance and call disconnect callback
        self.ws = None

        if self.on_disconnect:
            try:
                await self.on_disconnect()
            except Exception as e:
                self.logger.error(f"Disconnect callback failed: {e}")

    @tenacity.retry(
        stop=tenacity.stop_after_attempt(7),
        wait=tenacity.wait_exponential(multiplier=5, min=0, max=120),
        retry=tenacity.retry_if_exception_type(RetryableWebSocketException),
    )
    async def send_message(self, message: str) -> bool:
        """Send text message"""
        if not self.is_connected:
            self.logger.warning("Not connected, cannot send message")
            self.logger.info("Attempting to reconnect...")
            try:
                await self.connect()
            except Exception as e:
                self.logger.error(f"Failed to connect: {e}")
                raise RetryableWebSocketException() from e

        try:
            await self.ws.send(message)
            self.logger.debug(f"Sent: {message[:100]}{'...' if len(message) > 100 else ''}")
            return True
        except ConnectionClosed as e:
            self.logger.error(f"Connection closed, failed to send message: {e}")
            self.ws = None
            # when connection closed for unauthorized, raise exception and exit the agent
            if e.code == 3000:
                self.logger.info("Unauthorized, agent was removed by server")
                raise UnauthorizedError() from e
            raise RetryableWebSocketException() from e
        except Exception as e:
            self.logger.error(f"Failed to send message: {e}")
            raise e

    async def send_json(self, data: dict) -> bool:
        """Send JSON message"""
        # auto add sending_time
        if data.get("sending_time") is None:
            data["sending_time"] = datetime.now(UTC).isoformat()

        try:
            message = json.dumps(data, ensure_ascii=False)
            return await self.send_message(message)
        except Exception as e:
            self.logger.error(f"Send JSON failed: {e}")
            return False

    @tenacity.retry(
        # include the first time call
        stop=tenacity.stop_after_attempt(7),
        # timeline: [run]-0s-[run]-10s-[run]-20s-[run]-40s-[run]-80s-[run]-120s-[run]
        wait=tenacity.wait_exponential(multiplier=5, min=0, max=120),
        retry=tenacity.retry_if_exception_type(RetryableWebSocketException),
    )
    async def listen_loop(self) -> None:
        """Message listening loop"""
        self.logger.info("Starting WebSocket message loop...")

        if not self.is_connected:
            self.logger.error("No WebSocket connection to listen on!")
            self.logger.info("Attempting to reconnect...")
            try:
                await self.connect()
            except Exception as e:
                self.logger.error(f"Failed to connect: {e}")
                raise RetryableWebSocketException() from e

        try:
            self.logger.info("Entering message listening loop...")
            async for message in self.ws:
                self.logger.info(f"Received: {message[:100]}{'...' if len(message) > 100 else ''}")

                if self.on_receive_message:
                    try:
                        await self.on_receive_message(message, self)
                    except Exception as e:
                        self.logger.error(f"Message callback failed: {e}")

            # Listen loop exit with no error, maybe server closed the connection
            # try to reconnect
            logger.warning("Listen loop exit with no error, maybe server closed the connection, try to reconnect")
            raise RetryableWebSocketException()

        except RetryableWebSocketException:
            raise
        except ConnectionClosed as e:
            self.logger.warning(f"WebSocket connection closed by peer: {e}")
            # when connection closed for unauthorized, raise exception and exit the agent
            if e.code == 3000:
                self.logger.info("Unauthorized, agent was removed by server")
                raise UnauthorizedError() from e
            raise RetryableWebSocketException() from e
        except WebSocketException as e:
            self.logger.error(f"WebSocket error: {e}")
            if self.on_error:
                try:
                    await self.on_error(e)
                except Exception as callback_error:
                    self.logger.error(f"Error callback failed: {callback_error}")
            raise RetryableWebSocketException() from e
        except Exception as e:
            self.logger.error(f"Unexpected error in message loop: {e}")
            if self.on_error:
                try:
                    await self.on_error(e)
                except Exception as callback_error:
                    self.logger.error(f"Error callback failed: {callback_error}")
            # critical error, exit the agent and restart by systemd
            raise e
        finally:
            # Once the loop is exited, the connection is closed and can be retried
            self.ws = None
            self.logger.info("Message listening loop exit...")

    def set_callbacks(
        self,
        on_receive_message: Callable[[str, "WebSocketConnector"], Coroutine[Any, Any, None]] | None = None,
        on_connect: Callable[[], Coroutine[Any, Any, None]] | None = None,
        on_disconnect: Callable[[], Coroutine[Any, Any, None]] | None = None,
        on_error: Callable[[Exception], Coroutine[Any, Any, None]] | None = None,
    ) -> None:
        """Set callback functions"""
        if on_receive_message is not None:
            self.on_receive_message = on_receive_message
        if on_connect is not None:
            self.on_connect = on_connect
        if on_disconnect is not None:
            self.on_disconnect = on_disconnect
        if on_error is not None:
            self.on_error = on_error

    async def __aenter__(self):
        """Async context manager entry"""
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit"""
        await self.disconnect()



import asyncio
import os
from collections.abc import Sequence
from typing import Callable, TypeAlias

import aiofiles
from docker.models.containers import Container
from docker.models.images import Image
from loguru import logger

from recurvedata.agent._internal.client import AgentClient
from recurvedata.agent._internal.cube.push_config import cube_push_config_service
from recurvedata.agent._internal.cube.restart_service import cube_restart_service
from recurvedata.agent._internal.docker import docker_client
from recurvedata.agent.config import CONFIG, RECURVE_HOME

from .schemas import (
    Action,
    ContainerManagementPayload,
    CubePushConfigPayload,
    CubeRestartServicePayload,
    DeployServicePayload,
    DeployServiceResult,
    UndeployServicePayload,
    WorkerManagementPayload,
)

COMPOSE_FILE_NAME = os.path.join(RECURVE_HOME, "docker-compose.yml")
COMPOSE_PROJECT_NAME = "recurve_services"
CUBE_DIR = os.path.join(RECURVE_HOME, "cube")
CUBE_CONF_DIR = os.path.join(CUBE_DIR, "conf")

# Type aliases
TaskList: TypeAlias = Sequence[asyncio.Task[None]]


registry: dict[Action, Callable[[], None]] = {}


def register_action(action_type: Action):
    def decorator(func: Callable[[], None]) -> Callable[[], None]:
        registry[action_type] = func
        return func

    return decorator


def get_run_env() -> dict[str, str]:
    env = os.environ.copy()
    env["RECURVE_HOME"] = str(RECURVE_HOME)
    return env


class WorkerManagementTask:
    @staticmethod
    async def _create_cube_conf_dir(payload: DeployServicePayload) -> None:
        if not payload.cube_env:
            return

        if not os.path.exists(CUBE_CONF_DIR):
            os.makedirs(CUBE_CONF_DIR)

        if payload.cube_env:
            async with aiofiles.open(os.path.join(CUBE_CONF_DIR, ".env"), "w") as f:
                await f.write(payload.cube_env)

        if payload.cube_python:
            # async with aiofiles.open(os.path.join(CUBE_CONF_DIR, "cube.py"), "w") as f:
            # await f.write(payload.cube_python)
            client = AgentClient(CONFIG)
            response = await client.request("GET", "/api/cube/config/cube-python")
            if response.get("cube_python"):
                cube_python_file = response.get("cube_python")
            else:
                cube_python_file = payload.cube_python

            async with aiofiles.open(os.path.join(CUBE_CONF_DIR, "cube.py"), "w") as f:
                await f.write(cube_python_file)

    @staticmethod
    async def _write_cube_proxy_config(payload: DeployServicePayload) -> None:
        if not payload.cube_proxy_config:
            return

        tasks: list[asyncio.Task[None]] = []
        proxy_config = payload.cube_proxy_config
        if proxy_config and isinstance(proxy_config, dict):
            for key, value in proxy_config.items():
                tasks.append(asyncio.create_task(WorkerManagementTask._write_proxy_config_file(str(key), str(value))))
            await asyncio.gather(*tasks)

    @staticmethod
    async def _write_proxy_config_file(key: str, value: str) -> None:
        async with aiofiles.open(os.path.join(CUBE_DIR, f"{key}.toml"), "w") as f:
            await f.write(value)

    @staticmethod
    async def _write_docker_compose_file(payload: DeployServicePayload) -> None:
        async with aiofiles.open(COMPOSE_FILE_NAME, "w") as f:
            await f.write(payload.docker_compose)

    @staticmethod
    async def _fetch_local_image_version(worker_image: str) -> str:
        try:
            image: Image = await asyncio.to_thread(docker_client.images.get, worker_image)
            worker_version = image.labels.get("VERSION")
            return worker_version or "Unknown version"
        except Exception:
            return f"Image {worker_image} not found."

    @staticmethod
    @register_action(Action.DEPLOY)
    async def deploy_service(payload: DeployServicePayload) -> DeployServiceResult | None:
        tasks: list[asyncio.Task[None]] = [
            asyncio.create_task(WorkerManagementTask._create_cube_conf_dir(payload)),
            asyncio.create_task(WorkerManagementTask._write_cube_proxy_config(payload)),
            asyncio.create_task(WorkerManagementTask._write_docker_compose_file(payload)),
        ]
        await asyncio.gather(*tasks)

        env = get_run_env()

        docker_compose_cmd = [
            "docker",
            "compose",
            "-f",
            COMPOSE_FILE_NAME,
            "-p",
            COMPOSE_PROJECT_NAME,
            "up",
            "-d",
        ]
        try:
            process = await asyncio.create_subprocess_exec(
                *docker_compose_cmd,
                env=env,
            )
            stdout, stderr = await process.communicate()

            if process.returncode != 0:
                error_msg = f"Docker compose failed: {stderr or stdout}"
                logger.error(error_msg)
                raise RuntimeError(error_msg)
        except Exception as e:
            error_msg = f"Docker compose failed: {str(e)}"
            logger.error(error_msg)
            raise

        logger.info("Docker Compose started successfully.")

        worker_version = await WorkerManagementTask._fetch_local_image_version(payload.worker_image)
        return DeployServiceResult(worker_version=worker_version)

    @staticmethod
    @register_action(Action.UNDEPLOY)
    async def undeploy_service(payload: UndeployServicePayload) -> None:
        try:
            cmd = ["docker", "compose", "-f", COMPOSE_FILE_NAME, "-p", COMPOSE_PROJECT_NAME, "down"]
            if payload.remove_volumes:
                cmd.append("-v")

            env = get_run_env()

            process = await asyncio.create_subprocess_exec(
                *cmd,
                env=env,
            )
            _, stderr = await process.communicate()

            if process.returncode != 0:
                error_msg = f"Docker compose failed: {stderr.decode()}"
                logger.error(error_msg)
                # raise for retry
                raise RuntimeError(error_msg)

        except Exception as e:
            error_msg = f"Docker compose failed: {str(e)}"
            logger.error(error_msg)
            raise

        logger.info("Undeploy service successfully")

    @staticmethod
    @register_action(Action.START)
    async def start_container(payload: ContainerManagementPayload) -> None:
        for container_name in payload.container_names:
            try:
                container: Container = await asyncio.to_thread(docker_client.containers.get, container_name)
                await asyncio.to_thread(container.start)
                logger.info(f"Container {container_name} started successfully")
            except Exception as e:
                error_msg = f"Failed to start container {container_name}: {str(e)}"
                logger.error(error_msg)
                raise

    @staticmethod
    @register_action(Action.RESTART)
    async def restart_container(payload: ContainerManagementPayload) -> None:
        for container_name in payload.container_names:
            try:
                container: Container = await asyncio.to_thread(docker_client.containers.get, container_name)
                await asyncio.to_thread(container.restart)
                logger.info(f"Container {container_name} restarted successfully")
            except Exception as e:
                error_msg = f"Failed to restart container {container_name}: {str(e)}"
                logger.error(error_msg)
                raise

    @staticmethod
    @register_action(Action.STOP)
    async def stop_container(payload: ContainerManagementPayload) -> None:
        for container_name in payload.container_names:
            try:
                container: Container = await asyncio.to_thread(docker_client.containers.get, container_name)
                await asyncio.to_thread(container.stop)
                logger.info(f"Container {container_name} stopped successfully")
            except Exception as e:
                error_msg = f"Failed to stop container {container_name}: {str(e)}"
                logger.error(error_msg)
                raise

    @staticmethod
    @register_action(Action.CUBE_PUSH_CONFIG)
    async def push_cube_config(payload: CubePushConfigPayload) -> None:
        await cube_push_config_service.push_cube_config(payload)

    @staticmethod
    @register_action(Action.CUBE_RESTART_SERVICE)
    async def restart_cube_service(payload: CubeRestartServicePayload) -> None:
        await cube_restart_service.restart_cube_service(payload)

    @staticmethod
    async def handle(payload: WorkerManagementPayload) -> DeployServiceResult | None:
        return await registry[payload.action](payload.payload)
